# Agent class
# Arika Ligmann-Zielinska, 2010
# -----------------------------
import numpy, random
import drules
# -----------------------------
# constants for attitudes
attitudes = {1:"UNBIASED", 2:"RECKLESS", 3:"CAUTIOUS",
             4:"POOR", 5:"RICH"}

class Developer:
    
    # constructor
    def __init__(self,attitude,demand,\
                 landvalue_weight,shape):
        """ This function creates a Developer object
            INPUT -> attitude (str)
                     demand (int)
                     weight (float)
                     shape (tuple for rows and cols of the grid)
        """
        self.attitude = attitude
        self.drule = 1 # FIXED
        self.demand = demand
        self.sample = 1.0# 0.1 # FIXED
        self.landvalue_weight = landvalue_weight
        self.beauty_weight = 1.0 - landvalue_weight
        self.utilities = None
        self.investmentSet = None
        self.investmentPlan = None
        self.ownedLand = numpy.zeros(shape,int)
        self.purchased = 0 # purchased during current tick
        self.dissattisfaction = 0
        

    def scoreSites(self, grid):
        """ This function calculates the utilities of sites based on
            Ideal Point decision rule
        """
        self.utilities = drules.applyIdealPoint(self, grid)

    def createInvestmentSet(self,grid,developable = 0):
        """ This function chooses a sample of developable locations
            INPUT -> Space object, int (optional)
        """
        # randomly draws 'n' developable cells from landuse array
        # and assigns utility value for selected cells and -9999.0 for the rest
        # (i.e. non-developable, already developed, and non-selected)
        # NOTE: it is a utility array
        developable_grid = []
        for coords,cell in numpy.ndenumerate(grid.landuse):
            if cell == developable:
                developable_grid.append(coords)
        available = grid.devcount()
        samplesize = int(round(available*self.sample))
        selection = random.sample(developable_grid, samplesize)
        ini_grid = numpy.zeros(grid.shape, int)
        for coords in selection:
            row = coords[0]
            col = coords[1]
            ini_grid[row,col] = 1
        # NOTE: utilities cannot be 'None', they have to be calculated before
        investment = numpy.where(ini_grid == 1, self.utilities, -9999.0)
        self.investmentSet = investment

    def updateInvestmentSet(self,grid,developable = 0):
        """ This function updates the sample of developable locations
            with sites already developed that have to be excluded from consideration
            it then randomly adds developable sites to meet the required sample size
            INPUT -> Space object, int (optional)
        """
        # update investment with developed sites
        self.investmentSet = numpy.where(grid.landuse != developable, -9999.0, self.investmentSet)
        # extend investment set so that it meets the sample size
        currentsample = numpy.sum(numpy.where(self.investmentSet != -9999.0,1,0))
        available = grid.devcount()
        maxsample = int(round(available*self.sample))
        if currentsample < maxsample and self.demand - self.purchased > 0:
            # no need to increase sample if:
            # [1] current sample >= max possible sample for this developer
            # [2] demand is met (finished investment for this tick)
            addsample = maxsample - currentsample
            developable_grid = []
            for coords,cell in numpy.ndenumerate(grid.landuse):
                if cell == developable and\
                   self.investmentSet[coords[0],coords[1]] == -9999.0:
                    # make sure that the site is not already in investment set
                    developable_grid.append(coords)
            selection = random.sample(developable_grid, addsample)
            for coords in selection:
                row = coords[0]
                col = coords[1]
                self.investmentSet[row,col] = self.utilities[row,col]

    def createInvestmentPlan(self):
        """ This function ranks selected locations based on their utilities
            and then selects k demanded cells based on rank (from best to worst)
        """
        # create a list of tuples 't' where t[0] is utility and t[1] is a tuple
        # of indices (location)
        utilitylist = []
        for index, utility in numpy.ndenumerate(self.investmentSet):
            utilitylist.append((utility,index))
        # shuffle so that the cells with the same utilities are
        # 'randomly' ordered in the next step
        random.shuffle(utilitylist)
        # sort based on utility
        utilitylist.sort(cmp)
        ranking = numpy.zeros(self.utilities.shape,int)
        currentlydemanded = self.demand - self.purchased
        demandlist = []
        # pick demanded cells based on utility
        if currentlydemanded > 0: # demand not yet met
            demandlist = utilitylist[:currentlydemanded]
            rank = 1
            for utilitydata in demandlist:
                row = utilitydata[1][0]
                col = utilitydata[1][1]
                ranking[row,col] = rank
                rank += 1
        self.investmentPlan = ranking

    def clear(self):
        """ This function clears 'market assessment' attributes after investment
            and prepares the developer for the next tick
        """
        self.purchased = 0
        self.utilities = None
        self.investmentSet = None
        self.investmentPlan = None

# functions for a collection of developers

def bids(developers):
    """ This function compares pending 'investment plans'
        of developer and returns the best ranks for each site
        INPUT -> developer agents (list of Developers)
        OUTPUT -> 2D numpy array of best (min) ranks
    """
    # build a stack of investment plans
    plans = []
    for developer in developers:
        plans.append(developer.investmentPlan)
    plans = tuple(plans)
    planstack = numpy.dstack(plans)
    # create an array of best (min) ranks for cells
    # NOTE: the best rank is 1 but we have min=0 for non-ranked sites
    # the zeros have to be excluded from consideration by setting them
    # to the worst possible rank
    # procedure -> find the max value in stack
    #           -> add 1 to it
    #           -> convert all zeros to this value, making out-of-ranking
    #              cells the least preferred
    maxrank = planstack.max()+1   
    planstack = numpy.where(planstack == 0, maxrank, planstack)
    # best ranks
    minranks = planstack.min(axis=2)
    # go back from maxrank to zeros for non-ranked sites
    minranks = numpy.where(minranks == maxrank, 0, minranks)
    return minranks

def ties(developers, bestranks):
    """ This function counts the number of rank ties per site
        INPUT -> developer agents (list of Developers),
                 rank array (2D numpy of int)
        OUTPUT -> array of ties (2D numpy of int)
    """    
    # count ties
    tiecounts = numpy.zeros(bestranks.shape, int)
    for developer in developers:
        plan = developer.investmentPlan
        conflicts = (bestranks == plan)
        # exclude non-ranked sites from conflicts
        conflicts = numpy.where(bestranks == 0, 0, conflicts)
        tiecounts += conflicts
    return tiecounts

def buy(developers, bids, ties):
    """ This function assigns parcels to developers based on their ranks;
        ties are resolved randomly.
        INPUT -> developer agents (list of Developers),
                 rank array (2D numpy of int)
                 tie array (2D numpy of int)
    """
    # [1] assign non-ties
    for developer in developers:
        # agent's investment ranking is equal to highest bids and no ties
        purchase = numpy.logical_and(developer.investmentPlan == bids, ties == 1)
        developer.ownedLand += purchase
        developer.purchased += numpy.sum(purchase)
    # [2] resolve ties
    if numpy.any(ties > 1):
        tielocs = {}
        for loc,tie in numpy.ndenumerate(ties):
            if tie > 1:
                tielocs[loc]= [] # placeholder for developers
        # check ties for each developer and add its ID to tielocs dictionary
        for developer in developers:
            # agent's investment ranking is equal to highest bids and is in a tie
            devties = numpy.logical_and(developer.investmentPlan == bids, ties > 1)
            for loc, devtie in numpy.ndenumerate(devties):
                if devtie == 1:
                    tielocs[loc].append(id(developer))
        # draw a winner
        devsold = {}
        for developer in developers:
            # initialize developer's wins
            devsold[id(developer)] = []
        for item in tielocs.items():
            location = item[0]
            devs = item[1]
            winnerid = random.choice(devs)
            devsold[winnerid].append(location)
        # assign resolved ties
        for developer in developers:
            purchase = devsold[id(developer)]
            developer.purchased += len(purchase)
            for location in purchase:
                developer.ownedLand[location[0],location[1]] = 1
        # compute dissasisfaction
        for developer in developers:
            developer.dissattisfaction += (developer.demand - developer.purchased)

def stillneeded(developers):
    """ This function returns total non-satisfied demand for all developers
        INPUT -> developer agents (list of Developers)
        OUTPUT -> int
    """
    return sum([(developer.demand - developer.purchased) for \
                developer in developers])
# helpers
def cmp(cell,otherCell):
    """ This function compares cell objects based
        on their score attribute.
        INPUT -> two Cell objects
    """
    # the result is reverse order (from highest to lowest)
    if cell[0] < otherCell[0]:
        return 1
    elif cell[0] == otherCell[0]:
        return 0
    else:
        return -1
