# drules.py DECISION RULES FOR CALCULATING SITE UTILITY
# Arika Ligmann-Zielinska, 2007-2008
# updated to numpy
# **************************************************************
import numpy
# importing user-defined libraries
import auf

# defining symbolic constants
attitudes = {1:"UNBIASED", 2:"RECKLESS", 3:"CAUTIOUS",
             4:"POOR", 5:"RICH"}

def applyIdealPoint(decider, space):
    """ This function returns site utility grid based on
        an Ideal Point aggregation function.
        INPUT -> Agent (for weights and attitude)
                 Space (for criteria)
        OUTPUT -> 2D array of utilities
    """
    # NOTE: if a criterion layer is added, the algorithm
    # needs a direct update, it does not take an arbitrary number
    # of decision criteria
    #
    # calculate gains and losses
    landvalue_gain = space.landvalue - getNadir(space.landvalue,space.landuse)
    landvalue_loss = space.landvalue - getIdeal(space.landvalue,space.landuse)
    beauty_gain = space.beauty -\
                          getNadir(space.beauty,space.landuse)
    beauty_loss = space.beauty -\
                          getIdeal(space.beauty,space.landuse)
    # recalculate gains and losses based on attitude
    perception = decider.attitude
    if perception == "UNBIASED":
        perceived_landvalue_gain = auf.unbiased(landvalue_gain)
        perceived_landvalue_loss = auf.unbiased(landvalue_loss)
        perceived_beauty_gain = auf.unbiased(beauty_gain)
        perceived_beauty_loss = auf.unbiased(beauty_loss)
    elif perception == "RECKLESS":
        perceived_landvalue_gain = auf.reckless(landvalue_gain)
        perceived_landvalue_loss = auf.reckless(landvalue_loss)
        perceived_beauty_gain = auf.reckless(beauty_gain)
        perceived_beauty_loss = auf.reckless(beauty_loss)
    elif perception == "CAUTIOUS":
        perceived_landvalue_gain = auf.cautious(landvalue_gain)
        perceived_landvalue_loss = auf.cautious(landvalue_loss)
        perceived_beauty_gain = auf.cautious(beauty_gain)
        perceived_beauty_loss = auf.cautious(beauty_loss)
    elif perception == "POOR":
        perceived_landvalue_gain = auf.poor(landvalue_gain)
        perceived_landvalue_loss = auf.poor(landvalue_loss)
        perceived_beauty_gain = auf.poor(beauty_gain)
        perceived_beauty_loss = auf.poor(beauty_loss)
    elif perception == "RICH":
        perceived_landvalue_gain = auf.rich(landvalue_gain)
        perceived_landvalue_loss = auf.rich(landvalue_loss)
        perceived_beauty_gain = auf.rich(beauty_gain)
        perceived_beauty_loss = auf.rich(beauty_loss)
    else: # placeholder for other functions (future)
        print "incorrect attitude, setting to UNBIASED"
        perceived_landvalue_gain = auf.unbiased(landvalue_gain)
        perceived_landvalue_loss = auf.unbiased(landvalue_loss)
        perceived_beauty_gain = auf.unbiased(beauty_gain)
        perceived_beauty_loss = auf.unbiased(beauty_loss)
    # calculate separations
    separIdeal = numpy.power(decider.landvalue_weight*perceived_landvalue_loss,2) +\
                 numpy.power(decider.beauty_weight*perceived_beauty_loss,2)
    separIdeal = numpy.sqrt(separIdeal)
    separNadir = numpy.power(decider.landvalue_weight*perceived_landvalue_gain,2) +\
                 numpy.power(decider.beauty_weight*perceived_beauty_gain,2)
    separNadir = numpy.sqrt(separNadir)
    # calculate score
    # when division by 0 occurs, set utility to -1.0
    utilities = numpy.where(separIdeal+separNadir == 0.0,-1.0,\
                            separNadir/(separNadir + separIdeal))
    return utilities

# -------
# helpers

def getIdeal(criterion, developable, luse = 0):
    """ This function returns an ideal for input layer. It
        considers only those cells that currently have the
        input landuse 'luse'.
        
        INPUT -> 2D array (floats), 2D array (ints), int (optional)
        OUTPUT -> float
    """
    selection_grid = numpy.where(developable == luse, criterion, -9999)
    return numpy.max(selection_grid)

def getNadir(criterion, developable, luse = 0):
    """ This function returns a nadir for input layer. It
        considers only those cells that currently have the
        input landuse 'luse'.
        
        INPUT -> 2D array (floats), 2D array (ints), int (optional)
        OUTPUT -> float
    """
    selection_grid = numpy.where(developable == luse, criterion, 9999)
    return numpy.min(selection_grid)
