# Arika Ligmann-Zielinska, MSU, 2010
# 
# numpy-based
#*****************************************************
import numpy, math

def load(infile, format = 'f'):
    """ This function returns an array of data of input txt grid-like file.
        It uses numpy IO functionality.
        INPUT -> *.asc *.txt (data file)
                 char for format
        OUTPUT -> 2D Array of 'format':
                  'f' is float (default)
                  'i' is integer
    """
    outarray = None
    if format == 'f':
        outarray=numpy.loadtxt(infile, dtype = float)
    elif format == 'i':
        outarray=numpy.loadtxt(infile,dtype=int)
    else: # placeholder for other formats
        print "Load data: improper format, select 'f' for float or 'i' for integer"
    return outarray

def save(inarray, filename, format = 'i'):
    """ This function saves the body of input 2D array to a txt file.
        It uses numpy IO functionality.
        INPUT -> 2D numpy array
                 the name of file (string)
                 char for format
    """
    if format == 'f':
        # float will be save with max 4 digits float
        numpy.savetxt(filename,inarray,fmt ='%.6f',delimiter = ' ')
    elif format == 'i':
        numpy.savetxt(filename,inarray, fmt = '%d',delimiter = ' ')
    else: # placeholder for other formats
        print "Save data: improper format, select 'f' for float or 'i' for integer"
    #print "File",filename,"saved"
        
# ----------
class Space:
    
    def __init__(self,landvaluefile,scenicbeautyfile):
        self.beauty = load(scenicbeautyfile)
        self.landvalue = load(landvaluefile)
        self.shape = self.beauty.shape
        self.size = self.beauty.size
        self.landuse = numpy.zeros(self.shape)
        

    def __str__(self):
        description = "grid shape "+str(self.shape)+\
                      "; "+"grid size "+str(self.size)
        return description

    def devcount(self, developable = 0):
        """ This function counts the number of developable locations
            INPUT -> int (optional)
            OUTPUT -> int
        """
        devnum = numpy.sum(numpy.where(self.landuse == developable,1,0))
        if devnum == 0:
            print "CRITICAL CONSTRAINT: no more land to develop"
        return devnum

    def develop(self, developers, luse = 1):
        """ This function sets undeveloped cells to 'luse'
            based on land owned by input developers
            INPUT -> developers (list of agents)
                     developed landuse type (int, optional)
        """
        for developer in developers:
            self.landuse = numpy.where(developer.ownedLand == 1, luse, self.landuse)


    def decrease_beauty(self,v,developed=1):
        """ natural beauty of 3x3 neighbors (including self) of developed cells
            decreases by fraction v: new_val = old_val - old_val*v
        """
        # NOTE, what follows is that:
        # -> natural beauty decreases constantly for adjacent neighs of all dev sites
        #    and not only the sites newly developed
        # -> for sites that have more than one developed neighbor, their beauty
        #    decreases by fraction v*d where d in the number of developed neighbors
        #    i.e. the decrease is cumulative based on the level of neighborhood development
        max_row = self.shape[0]-1
        max_col = self.shape[1]-1
        for index,cell in numpy.ndenumerate(self.landuse):
            if cell == developed:
                neighs = getNeighborIndices(index,max_row,max_col)
                neighs.append(index)
                for neigh in neighs:
                    delta = self.beauty[neigh[0],neigh[1]]*v
                    if self.beauty[neigh[0],neigh[1]]-delta >= 0.0:
                        self.beauty[neigh[0],neigh[1]] -= delta
                    else:
                        # new value cannot be less than 0.0 (normalized layer)
                        self.beauty[neigh[0],neigh[1]] = 0.0

    def increase_landvalue(self,f,developed=1):
        """ vals of 3x3 neighbors (including self) of a developed cell (d_landval)
            increase by fraction f: new_val = f*d_landval + old_val
            thus new value is directly influenced by the value development
        """
        # NOTE, what follows is that:
        # -> land value increases constantly for adjacent neighs of all dev sites
        #    and not only the sites newly developed
        # -> for sites that have more than one developed neighbor, their land value
        #    increases by fraction f*landvalue for each developed neighbor
        #    i.e. the increase is cumulative based on the level of neighborhood development
        max_row = self.shape[0]-1
        max_col = self.shape[1]-1
        for index,cell in numpy.ndenumerate(self.landuse):
            if cell == developed:
                delta = f*self.landvalue[index[0],index[1]]
                neighs = getNeighborIndices(index,max_row,max_col)
                neighs.append(index)
                for neigh in neighs:
                    if self.landvalue[neigh[0],neigh[1]]+delta < 1.0:
                        self.landvalue[neigh[0],neigh[1]] += delta
                    else:
                        # new value cannot exceed 1.0 (normalized layer)
                        self.landvalue[neigh[0],neigh[1]] = 1.0
    def AI(self, classval=1):
        """ returns Aggregation Index of the current landuse map """
        # specification:
        # http://www.umass.edu/landeco/research/fragstats/documents/Metrics/Contagion%20-%20Interspersion%20Metrics/Metrics/C116%20-%20AI.htm
        # defaults: the land use class is 1
        # NOTE: comparison with Fragstats gave results that were
        # [1] lower in absolute terms but
        # [2] relative difference in AI between 2 maps was similar
        maxrow = self.shape[0]-1
        maxcol = self.shape[1]-1
        devcount = self.landuse.sum()
        if devcount <= 1:
            return "N/A" # AI is not applicable for one cell 
        n = int(math.sqrt(devcount))
        m = devcount - n*n
        maxg = None
        if m == 0: maxg = 2*n*(n-1)
        elif m <= n: maxg = 2*n*(n-1) + 2*m - 1
        else: maxg = 2*n*(n-1) + 2*m - 2
        rowscount = self.shape[0]
        colscount = self.shape[1]
        adjcount = 0
        # count adjacent
        for row in range(rowscount):
            for col in range(colscount):
                leftcol = col + 1
                downrow = row + 1
                if leftcol < colscount-1:
                    if self.landuse[row,col] == self.landuse[row,leftcol] == classval:
                        adjcount += 1
                if downrow < rowscount-1:
                    if self.landuse[row,col] == self.landuse[downrow,col] == classval:
                        adjcount += 1
        AI = float(adjcount)/maxg*100
        return AI

# %%%%%%%
# helpers

def getNeighborIndices(location,rows,cols):
    """(tuple,int,int) -> a list of tuples of coordinates"""
    # returns the indices of the 3x3 neighbors of the input location
    row = location[0]
    col = location[1]
    # neighbor indices    
    N = (row-1,col)
    NW = (row-1,col+1)
    W = (row,col+1)
    SW =(row+1,col+1)
    S = (row+1,col)
    SE = (row+1,col-1)
    E = (row,col-1)
    NE = (row-1,col-1)
    indices = [N,NW,W,SW,S,SE,E,NE]
    neighbors = []
    # make sure that the indices are not out of bounds
    for coords in indices:
        nrow = coords[0]
        ncol = coords[1]
        if 0 <= nrow <= rows and 0 <= ncol <= cols:
            neighbors.append(coords)
    return neighbors
